#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <thread>
#include <chrono>
#include <ctime>

#include "workflow/HttpMessage.h"
#include "workflow/WFTaskFactory.h"
#include "workflow/WFFacilities.h"
#include "workflow/Workflow.h"

#include "client_config/config.h"

static srpc::ClientRPCConfig config;

struct MMInput
{
    int id;
	std::string file_path;
};

struct MMOutput
{
    int error;
	double time;
};

using MMFactory = WFThreadTaskFactory<MMInput, MMOutput>;
using MMTask = WFThreadTask<MMInput, MMOutput>;

void put_callback(WFHttpTask *task)
{
    int state = task->get_state();
    int error = task->get_error();

    if (state == WFT_STATE_SUCCESS) // print server response body
    {
        const void *body;
        size_t body_len;

        task->get_resp()->get_parsed_body(&body, &body_len);
        std::string res((char *)(body), body_len);
        if (res.empty())
        {
            return;
        }
        auto json_res = wfrest::Json::parse(res);
        config.add_http_put_count();
    }
}

void convert_callback(MMTask *task)
{
	auto *input = task->get_input();
	auto *output = task->get_output();

    std::string url = std::string("http://") + config.client_host() +
                      std::string(":") + std::to_string(config.client_port()) + std::string("/UpdateTaskStatus");
    std::stringstream query_params;
    query_params << "?id=" << input->id << "&status=";
    if (output->error == 0) {
        query_params << 2;
    } else {
        query_params << 3;
    }
    
    url.append(query_params.str());
	assert(task->get_state() == WFT_STATE_SUCCESS);

    WFHttpTask *http_task = WFTaskFactory::create_http_task(url,
                                                            config.redirect_max(),
                                                            config.retry_max(),
                                                            put_callback);
    http_task->start();
}

void convert_pdf(const MMInput *in, MMOutput *out)
{
    out->error = -1;
    if (!in->file_path.empty())
    {
        char* new_conv_args[7] {nullptr};
        for (int i = 0; i < 6; ++i) {
            switch (i)
            {
            case 0:
            {
                new_conv_args[i] = new char[strlen(config.get_conversion_instance()) + 1]{0};
                strcpy(new_conv_args[i], config.get_conversion_instance());
                break;
            }
            case 1:
            {
                new_conv_args[i] = new char[strlen(config.get_conversion_type()) + 1]{0};
                strcpy(new_conv_args[i], config.get_conversion_type());
                break;
            }
            case 2:
            {
                new_conv_args[i] = new char[strlen(config.get_conversion_resource_path()) + 1]{0};
                strcpy(new_conv_args[i], config.get_conversion_resource_path());
                break;
            }
            case 3:
            {
                new_conv_args[i] = new char[in->file_path.length() + 1]{0};
                strcpy(new_conv_args[i], in->file_path.c_str());
                break;
            }
            case 4:
            {
                std::string output_path = config.get_output_path();
                output_path.append(config.get_time_stamp());
                output_path.append("/");

                std::string tmp_str = in->file_path;
                tmp_str.erase(0, strlen(config.get_root_path()));
                auto position = tmp_str.find_last_of('/');
                tmp_str.erase(position, tmp_str.length() - position);

                output_path.append(tmp_str);
                output_path.append("/");
                new_conv_args[i] = new char[output_path.length() + 1]{0};
                strcpy(new_conv_args[i], output_path.c_str());
                break;
            }
            case 5:
            {
                new_conv_args[i] = new char[strlen(config.get_model_path()) + 1]{0};
                strcpy(new_conv_args[i], config.get_model_path());
                break;
            }
            default:
                break;
            }
        }

        out->error = 0;
        pid_t pid = fork();
        if (pid == -1) { // fork失败
            std::cerr << "Failed to fork()" << std::endl;
            return;
        } else if (pid == 0) { // 这里是子进程
            execvp(new_conv_args[0], new_conv_args);
            out->error = -1;
            // 此处不应该被执行，除非execvp失败
            _exit(EXIT_FAILURE); // 如果execvp失败，退出子进程
        } else { // 这里是父进程
            int status;
            // 等待子进程结束，并获取子进程的退出状态
            if (waitpid(pid, &status, 0) == -1) {
                perror("waitpid");
                out->error = -1;
                return;
            }

            // 检查子进程的退出状态
            if (WIFEXITED(status)) {
                out->error = WEXITSTATUS(status);
            } else if (WIFSIGNALED(status)) {
                std::cerr << "Child process terminated by signal." << std::endl;
                out->error = -1; // 或者设置一个特定的错误码
            }
        }

        for (int i = 0; i < 7; ++i)
            delete[] new_conv_args[i];
    }
}

void sig_handler(int signo)
{
    
}

void init(const char* path)
{
    if (config.load(path) == false)
    {
        perror("Load config failed");
        exit(1);
    }

    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);
}

void callback(WFHttpTask *task)
{
    int state = task->get_state();
    int error = task->get_error();

    if (state == WFT_STATE_SUCCESS) // print server response body
    {
        const void *body;
        size_t body_len;

        task->get_resp()->get_parsed_body(&body, &body_len);
        std::string res((char *)(body), body_len);
        if (res.empty())
        {
            return;
        }
        
        auto json_res = wfrest::Json::parse(res);

        if (json_res.has("path"))
        {
            std::string file_path = config.get_pdf_root_path() + json_res["path"].get<std::string>();
            MMTask *task = MMFactory::create_thread_task("convert_task", convert_pdf, convert_callback);
            auto *input = task->get_input();
            input->file_path = std::move(file_path);
            input->id = json_res["id"].get<int>();
           
            Workflow::start_series_work(task, nullptr);
            config.add_http_get_count();
        }
    }
}

int main(int argc, char* argv[])
{
    if (argc < 2)
    {
        std::cout << "please input client config path." << std::endl;
        return -1;
    }
    

    init(argv[1]);

    std::string url = std::string("http://") + config.client_host() +
                      std::string(":") + std::to_string(config.client_port()) + std::string("/GetNewTask");

    auto outer_size = config.get_outer_batch_size();
    auto inner_size = config.get_inner_batch_size();
    for (int i = 0; i < outer_size; i++)
    {
        auto begin = std::chrono::high_resolution_clock::now();
        ParallelWork *pwork = Workflow::create_parallel_work(nullptr);
        for (int j = 0; j < inner_size; j++)
        {
            WFHttpTask *task = WFTaskFactory::create_http_task(url,
                                                            config.redirect_max(),
                                                            config.retry_max(),
                                                            callback);
            SeriesWork *series = Workflow::create_series_work(task, nullptr);
            pwork->add_series(series);
        }
        
        Workflow::start_series_work(pwork, nullptr);
        auto end = std::chrono::high_resolution_clock::now();
        auto elapsed = std::chrono::duration_cast<std::chrono::nanoseconds>(end - begin);
        std::cout << "转换任务数: " << config.get_http_get_count() << ", 上传任务数: " << config.get_http_put_count() << std::endl;
    }
    
    return 0;
}