#include <stdio.h>
#include <signal.h>
#include <mutex>
#include <thread>

#include "workflow/WFFacilities.h"
#include "srpc/rpc_types.h"
#include "fastdeploy/vision.h"
#include "rapidjson/document.h"
#include "rapidjson/writer.h"
#include "rapidjson/stringbuffer.h"

#include "server_config/config.h"
#include "server.srpc.h"

using namespace srpc;

static int gpu_id = 0;

static std::vector< fastdeploy::vision::detection::PPYOLOER* > det_models;
static std::vector< char > model_used;
static std::mutex model_mtx;

static WFFacilities::WaitGroup wait_group(100);
static srpc::ServerRPCConfig config;

std::string DetectionResultToJson(const fastdeploy::vision::DetectionResult& res)
{
    rapidjson::StringBuffer ss;
    rapidjson::Writer<rapidjson::StringBuffer> writer(ss);

    auto label_map = [](int label) -> std::string {
        switch (label)
        {
            case 3:
                return "Figure";

            case 6:
                return "Table_0";

            case 7:
                return "Table_std";

            default:
                return "";
        }
    };

    writer.StartObject();
    writer.Key("objects");
    writer.StartArray();
    for (int i = 0; i < res.scores.size(); i++) {
        auto label_str = label_map(res.label_ids[i]);
        if (label_str.empty() || res.scores[i] < 0.5)
            continue;

        writer.StartObject();
        writer.Key("text");
        writer.String(label_str.c_str());

        writer.Key("score");
        writer.Double(res.scores[i]);

        writer.Key("position");
        writer.StartArray();
        writer.Int(res.boxes[i][0]);
        writer.Int(res.boxes[i][1]);
        writer.Int(res.boxes[i][2]);
        writer.Int(res.boxes[i][1]);
        writer.Int(res.boxes[i][2]);
        writer.Int(res.boxes[i][3]);
        writer.Int(res.boxes[i][0]);
        writer.Int(res.boxes[i][3]);
        writer.EndArray();
        writer.EndObject();
    }
    writer.EndArray();
    writer.EndObject();
    return ss.GetString();
}


void sig_handler(int signo)
{
    wait_group.done();
}

void init(const char* path)
{
    if (config.load(path) == false)
    {
        perror("Load config failed");
        exit(1);
    }

    fastdeploy::SetLogger(false, false);

    auto option = fastdeploy::RuntimeOption();
    option.UseGpu(gpu_id);

    int model_count = config.get_model_count();
    for(size_t i = 0; i < model_count; ++i) {
        if (i > model_count / 2 && gpu_id == 0)
        {
            gpu_id = 1;
            option.UseGpu(gpu_id);
            auto* model =  new fastdeploy::vision::detection::PPYOLOER(config.get_det_model_path(), config.get_det_params_path(), config.get_det_cfg_path(), option);
            model->Initialized();
            det_models.emplace_back(model);
            continue;
        }
        
        if (det_models.empty()) {
            auto* model =  new fastdeploy::vision::detection::PPYOLOER(config.get_det_model_path(), config.get_det_params_path(), config.get_det_cfg_path(), option);
            model->Initialized();
            det_models.emplace_back(model);
        } else {
            auto tmodel = std::move(det_models.back()->Clone());
            det_models.emplace_back(reinterpret_cast<fastdeploy::vision::detection::PPYOLOER*>(tmodel.release()));
        }
    }
    std::cout << "成功初始化 " << model_count << " 个模型！" << std::endl;
    model_used.resize(model_count, 0);

    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);
}

class ServiceImpl : public gpu_server::Service
{
public:
    void Echo(EchoRequest *req, EchoResponse *resp, RPCContext *ctx) override
    {
        try {
            auto get_valid_model = []() -> int {
                while (true) {
                    for (int i = 0; i < det_models.size(); ++i) {
                        if (!model_used[i]) {
                            std::unique_lock<std::mutex> lk(model_mtx);
                            if (!model_used[i])
                            {
                                model_used[i] = 1;
                                std::cout << "将使用第 " << i + 1 << " 个模型进行推理！" << std::endl;
                                return i;
                            }                            
                        }
                    }
                }
            };

            const cv::Mat im = cv::imread(req->path());
            if (!im.empty())
            {
                auto begin = std::chrono::high_resolution_clock::now();
                auto model_index = get_valid_model();

                fastdeploy::vision::DetectionResult res;
                if (!det_models[model_index]->Predict(im, &res)) {
                    std::cerr << "Failed to predict." << std::endl;
                    return;
                }
                // 4. delete the following codes and fill your logic
                {
                    std::lock_guard<std::mutex> gd(model_mtx);
                    model_used[model_index] = 0;
                }
                auto json_str = DetectionResultToJson(res);
                res.Free();
                resp->set_json(json_str.c_str());
                auto end = std::chrono::high_resolution_clock::now();
                auto elapsed = std::chrono::duration_cast<std::chrono::nanoseconds>(end - begin);
                // std::cout << "推理耗时: << " << elapsed.count() * 1e-9 << " seconds.\n";
            } else {
                resp->set_json("{}");
            }
        } catch (const std::exception& e){
            std::cout << "failed: " << e.what() << std::endl;
        }
    }
};

int main(int argc, char* argv[])
{
    if (argc < 2)
    {
        std::cout << "please input client config path." << std::endl;
        return -1;
    }

    // 1. load config
    init(argv[1]);

    // 2. start server
    SRPCServer server;
    ServiceImpl impl;
    server.add_service(&impl);

    config.load_filter(server);

    if (server.start(config.server_port()) == 0)
    {
        // 3. success and wait
        fprintf(stderr, "gpu_server SRPC server started, port %u\n", config.server_port());
        wait_group.wait();
        server.stop();
    }
    else
        perror("server start");

    return 0;
}